(( TreeView.F   A rudimentary TreeView class              by Michael Hillerstrm
                                                     michael-hillerstrom@usa.net


                This TreeView class hooks into Windows own library class.
                But be warned; this is a very 'stripped to the bone implementation'
                i.e. it has just what I need for DiaEdit...  Some day (soon) I
                will try to correct this.

                Please note that this code needs a new version of WINCON.DLL
                (dated September 15, 1997 or later).

                An example is included last in this file...


                Any comments/suggestions to:    michael-hillerstrom@usa.net



        Change log:
                November 23rd, 1997 - 21:59 MIH
                Added the line:   WinLibrary COMCTL32.DLL   to this source.

                September 16th, 1997 - 21:42 MIH
                Removed reference to COMMCTRL.F as Tom Zimmer has released an
                extended WINCON.DLL.  Thanks, Tom!

                August 31st, 1997 - 23:15 MIH
                First attempt...(which is VERY bare bones...)
                Need to convince Tom Zimmer to include #define's from COMMCTRL.H
                in WINCON.DLL.  For the time beeing, we'll have to cope with my
                FORTH constants in COMMCTRL.F.

))

anew TreeView.F

cr .( Loading TreeView Class...)

INTERNAL

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\       Prerequisites...
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

WinLibrary COMCTL32.DLL         \ Make sure that ComCtl32.dll is loaded...


\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\       TreeView Constants and their significance...
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

comment:

A treeview can have a combination of the following styles:


TVS_DISABLEDRAGDROP  disables drag n' drop of tree-view items.

TVS_LINESATROOT      draws lines linking child items to the root of the hierarchy.

TVS_HASLINES         enhances the graphic representation of a tree-view by
                     drawing lines that link child items to their parent item.
                     To link items at the root of the hierarchy, you need to
                     combine this and the TVS_LINESATROOT style.

TVS_HASBUTTONS       adds a button to the left side of each parent item. The
                     user can click the button to expand or collapse the child
                     items as an alternative to double-clicking the parent
                     item. To add buttons to items at the root of the
                     hierarchy, this style must be combined with TVS_HASLINES,
                     and TVS_LINESATROOT.

TVS_EDITLABELS       makes it possible for the user to edit the labels of
                     tree-view items.

TVS_SHOWSELALWAYS    causes a selected item to remain selected when the
                     tree-view control loses focus.


comment;

EXTERNAL

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\       TreeView Class...
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

:Class TreeViewControl <super Child-Window

     Record:    nmhdr
        int     hWndFrom
        int     idFrom
        int     code
    ;RecordSize: /nmhdr

     Record:    tvitem
        int     mask
        int     hItem
        int     state
        int     stateMask
        int     pszText
        int     cchTextMax
        int     iImage
        int     iSelectedImage
        int     cChildren
        int     lParam
    ;RecordSize: /tvitem

     Record:    tvins
        int     hParent
        int     hInsertAfter
\   TV_ITEM     item
/tvitem bytes   item
    ;RecordSize: /tvins

\    Record:    tvdi
\     NMHDR     hdr
\   TV_ITEM     item
\   ;Record


     Record:    tvkd
\     NMHDR     hdr
        int     wVKey
        int     flags
    ;RecordSize: /tvkd


     Record:    nmtv
\     NMHDR     hdr
        int     action
\   TV_ITEM     itemOld
        int     maskOld
        int     hItemOld
        int     stateOld
        int     stateMaskOld
        int     pszTextOld
        int     cchTextMaxOld
        int     iImageOld
        int     iSelectedImageOld
        int     cChildrenOld
        int     lParamOld
\   TV_ITEM     itemNew
        int     maskNew
        int     hItemNew
        int     stateNew
        int     stateMaskNew
        int     pszTextNew
        int     cchTextMaxNew
        int     iImageNew
        int     iSelectedImageNew
        int     cChildrenNew
        int     lParamNew
\     POINT     ptDrag
        int     x
        int     y
    ;RecordSize: /nmtv


: fill-nmhdr    ( l -- )
                abs>rel
                nmhdr   /nmhdr   2dup erase
                                      move
                ;
: fill-tvkd     ( l -- )
                abs>rel 3 cells+
                tvkd    /tvkd    2dup erase
                                      move
                ;
: fill-tvitem   ( l -- )
                abs>rel 3 cells+
                tvitem  /tvitem  2dup erase
                                      move
                ;

: fill-nmtv     ( addr -- )
                abs>rel 3 cells+
                nmtv    /nmtv    2dup erase
                                      move
                ;

: tvitem->tvins ( -- )
                tvitem item /tvitem   move ;


\ -------------------- Create Tree-View Control --------------------

create treeview-class   z," SysTreeView32"      \ Pre-registered class

: create-treeview ( -- hWnd )
                \ Make sure Common Controls are loaded
                Call InitCommonControls drop

                NULL                            \ Creation parameter
                appInst                         \ Instance handle
                id                              \ Child id
                Parent conhndl =
                if      conhndl
                else    GetHandle: Parent       \ parent window handle
                then
                tempRect.AddrOf GetClientRect: Parent
                tempRect.Bottom tempRect.Right  \ Size h,w
                0 0                             \ Position y,x
                WindowStyle: [ self ]           \ Style
                NULL                            \ Window name
                treeview-class rel>abs          \ Pre-registered class
                0                               \ Extended style
                Call CreateWindowEx
                ;


:M WindowStyle: ( -- style )
                WS_CHILD
                WS_VISIBLE          or
                ;M

:M Start:       ( Parent -- )
                hWnd
                if      drop
                        SW_SHOWNOACTIVATE Show: self
                else
                        to Parent
                        create-treeview to hWnd
                then
                ;M

:M AutoSize:    ( -- )
                tempRect.AddrOf GetClientRect: Parent
                0 0 tempRect.Right tempRect.Bottom   \ x,y,h,w
                Move: self
                ;M

:M Handle_Notify: ( h m w l -- f )
                dup fill-nmhdr
                code
                case
                   TVN_BEGINDRAGA      of  fill-nmtv   On_BeginDrag:      [ self ] endof
                   TVN_BEGINRDRAGA     of  fill-nmtv   On_BeginRDrag:     [ self ] endof
                   TVN_BEGINLABELEDITA of  fill-tvitem On_BeginLabelEdit: [ self ] endof
                   TVN_DELETEITEMA     of  fill-nmtv   On_DeleteItem:     [ self ] endof
                   TVN_ENDLABELEDITA   of  fill-tvitem On_EndLabelEdit:   [ self ] endof
                   TVN_GETDISPINFOA    of  fill-tvitem On_GetDispInfo:    [ self ] endof
                   TVN_ITEMEXPANDEDA   of  fill-nmtv   On_ItemExpanded:   [ self ] endof
                   TVN_ITEMEXPANDINGA  of  fill-nmtv   On_ItemExpanding:  [ self ] endof
                   TVN_KEYDOWN         of  fill-tvkd   On_KeyDown:        [ self ] endof
                   TVN_SELCHANGEDA     of  fill-nmtv   On_SelChanged:     [ self ] endof
                   TVN_SELCHANGINGA    of  fill-nmtv   On_SelChanging:    [ self ] endof
                   TVN_SETDISPINFOA    of  fill-tvitem On_SetDispInfo:    [ self ] endof
                          false swap ( default)
                endcase
                ;M


:M ToggleExpandItem:  ( hItem -- )
                TVE_TOGGLE TVM_EXPAND hWnd Call SendMessage drop
                ;M

:M SortChildren:      ( hItem -- )
                false TVM_SORTCHILDREN hWnd Call SendMessage drop
                ;M

\ --------------------- Overridable methods ----------------------

:M On_BeginDrag: ( -- f )
                false
                ;M
:M On_BeginRDrag: ( -- f )
                false
                ;M
:M On_BeginLabelEdit: ( -- f )  \ f=true, cancel edit,  f=false, ok edit
                false
                ;M
:M On_DeleteItem: ( -- f )
                false
                ;M
:M On_EndLabelEdit: ( -- f )
                false
                ;M
:M On_GetDispInfo: ( -- f )
                false
                ;M
:M On_ItemExpanded: ( -- f )
                false
                ;M
:M On_ItemExpanding: ( -- f )   \ f=true, don't expand/collapse, f=false, ok go ahead
                false
                ;M
:M On_KeyDown: ( -- f )
                false
                ;M
:M On_SelChanged: ( -- f )
                false
                ;M
:M On_SelChanging: ( -- f )     \ f=true, don't change, f=false, ok change
                false
                ;M
:M On_SetDispInfo: ( -- f )
                false
                ;M
;Class

MODULE


\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\       End  of Class definition(s)
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\


\   ***************************************************
\S  ******** Below please find a small example ********
\   ***************************************************


:Class NewTVC <super TreeViewControl


:M WindowStyle: ( -- style )
                WindowStyle: super
                TVS_HASLINES        or
                TVS_HASBUTTONS      or
                TVS_DISABLEDRAGDROP or
                TVS_SHOWSELALWAYS   or
                TVS_LINESATROOT     or
                ;M

int hRoot
int hSon
int hPrev

: AddItem       ( sztext hAfter hParent nChildren -- )
                tvins  /tvins  erase
                tvitem /tvitem erase
                ( nChildren)      to cChildren
                ( hParent)        to hParent
                ( hAfter)         to hInsertAfter
                rel>abs           to pszText
                TVIF_TEXT TVIF_CHILDREN or
                to mask
                tvitem->tvins
                tvins rel>abs 0 TVM_INSERTITEMA hWnd Call SendMessage
                to hPrev
                ;


: FillTreeView  ( -- )
                z" Root Item"        TVI_LAST  TVI_ROOT  1  AddItem
                hPrev to hRoot
                z" First son"        hPrev     hRoot     1  AddItem
                hPrev to hSon
                z" First Grandson"   hPrev     hSon      0  AddItem
                z" Second Grandson"  hPrev     hSon      0  AddItem
                z" Second son"       hPrev     hRoot     0  AddItem
                z" Third son"        hPrev     hRoot     1  AddItem
                hPrev to hSon
                z" Third Grandson"   hPrev     hSon      0  AddItem
                ;

:M Start:       ( Parent -- )
                Start: super

                \ Insert items...
                FillTreeView
                ;M


:M On_SelChanged: { \ text$ -- f }      \ Show text of selected item in message box
                maxstring LocalAlloc: text$

                TVIF_TEXT        to mask
                hItemNew         to hItem
                text$ rel>abs    to pszText
                maxstring        to cchTextMax
                tvitem rel>abs 0 TVM_GETITEMA  hWnd Call SendMessage drop

                text$ z" TreeView selection"
                MB_OK MessageBox: Parent drop

                false
                ;M

;Class


:Object MainWindow  <super Window

        int TreeView

:M On_Init:     ( -- )
                On_Init: super

                New> NewTVC to TreeView
                1001 SetId: TreeView
                self Start: TreeView
                ;M

:M WindowTitle: ( -- sztitle )
                z" TreeView Test"
                ;M

:M StartSize:   ( -- w h )
                100 100
                ;M

:M On_Size:     ( -- )
                AutoSize: TreeView
                ;M

:M WM_NOTIFY    ( h m w l -- f )
                dup abs>rel @  GetHandle: TreeView  =
                if      Handle_Notify: TreeView
                else    false
                then
                ;M

;Object


\ debug breaker

Start: MainWindow


\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\       End of File
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
